#!/usr/bin/perl

use Socket;
use IPC::Open2;
use POSIX ":sys_wait_h";

my $done_download = 0;
system('stty raw');

my $addr = shift || &usage;
my $port = shift || 23;
if ($port =~ /\D/)  {
	# Get service name if not a number
	$port = getservbyname($port, 'tcp')
}
&usage unless $port;
my $remote = sockaddr_in($port, inet_aton($addr));
socket(TELNET,PF_INET,SOCK_STREAM,getprotobyname('tcp')) || die "Socket: $!";
connect(TELNET,$remote);

# Main console I/O Loop.
while(1)  {
	my $rin='';
	vec($rin,fileno(TELNET),1) = 1;
	my $ein = $rin;
	vec($rin,fileno(STDIN),1) = 1;
	my $eout='';
	select($rin,undef,$eout=$ein,1);
	my $rin='';
	vec($rin,fileno(TELNET),1) = 1;
	if(select($rin,undef,undef,0))  {
		if (sysread(TELNET,my $rdata,65535))  {
			&input($rdata,*STDOUT);
		}
		else {
			&lost_carrier();
		}
	}
	my $rin='';
	vec($rin,fileno(STDIN),1) = 1;
	if(select($rin,undef,undef,0))  {
		if (sysread(STDIN,my $ldata,65535)) {
			if ($ldata =~ /^(.*)\x01(.*)$/s)  {
				&output($1);
				&external;
			}
			else  {
				&output($ldata);
			}
		}
	}
}

sub BEGIN  {
# Strip IAC escapes
# and play dumb (don't agree to anything!)
	my $in_iac=0;
	my $in_do=0;
	my $in_will=0;
	sub input  {
		my $instr = shift;
		my $outhandle = shift;
		my $writeout='';
		$instr=reverse $instr;
		while ((my $got=chop $instr) ne '')  {
			if($in_iac)  {
				$in_iac = 0;
				$got = "\xff".$got;
			}
			if($in_do)  {
				$in_do=0;
				syswrite(TELNET,"\xff\xfc".$got);
			}
			elsif($in_will)  {
				$in_will=0;
				syswrite(TELNET,"\xff\xfe".$got);
			}
			elsif($got eq "\xff")  {
				$in_iac=1;
			}
			elsif($got eq "\xff\xfd")  {
				$in_do=1;
			}
			elsif($got eq "\xff\xfb")  {
				$in_will=1;
			}
			elsif($got eq "\xff\xff")  {
				$writeout .= "\xff";
			}
			elsif(substr($got,0,1) ne "\xff")  {
				$writeout .= $got;
			}
		}
		syswrite($outhandle,$writeout) if ($writeout ne '');
	}
}

sub output  {
	my $send = shift;
	$send =~ s/\xff/\xff\xff/gs;
	syswrite(TELNET,$send);
}

sub usage  {
	print "Usage: $0 addr [<port>]\n";
	exit;
}

sub external  {
	system('stty sane');
	print "\nCommand: ";
	$command = <STDIN>;
	open2(*INB,*OUTB,$command) || return;
	while (!waitpid(-1,&WNOHANG))  {
		my $rin='';
		vec($rin,fileno(INB),1) = 1;
		vec($rin,fileno(TELNET),1) = 1;
		select($rin,undef,undef,.25);
		my $rin='';
		vec($rin,fileno(TELNET),1) = 1;
		if(select($rin,undef,undef,0))  {
			if (sysread(TELNET,my $rdata,65535))  {
				&input($rdata,*OUTB);
			}
			else  {
				close OUTB;
				close INB;
				system('stty raw');
				return;
			}
		}
		my $rin='';
		vec($rin,fileno(INB),1) = 1;
		if(select($rin,undef,undef,0))  {
			if (sysread(INB,my $cdata,65535)) {
				&output($cdata);
			}
			else  {
				close OUTB;
				close INB;
				system('stty raw');
				return;
			}
		}
	}
	close OUTB;
	close INB;
	system('stty raw');
}

sub lost_carrier  {
	system('stty raw');
	print "\n\nCarrier Lost\r\n";
	exit;
}
